/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 *
 *  Copyright © 2006 Pierre Habouzit
 */

/*
 * Copyright notice from original mutt:
 * Copyright (C) 1996-2000 Michael R. Elkins <me@mutt.org>
 * Copyright (C) 2000-2001 Edmund Grimley Evans <edmundo@rano.org>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#include <lib-lib/mem.h>
#include <lib-lib/str.h>
#include <lib-lib/ascii.h>

#include <lib-mime/mime.h>

#include "charset.h"
#include "thread.h"

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* If you are debugging this file, comment out the following line. */
/*#define NDEBUG*/

#ifdef NDEBUG
#define assert(x)
#else
#endif

#define ENCWORD_LEN_MAX 75
#define ENCWORD_LEN_MIN 9       /* m_strlen("=?.?.?.?=") */

#define HSPACE(x) ((x) == '\0' || (x) == ' ' || (x) == '\t')
#define CONTINUATION_BYTE(c) (((c) & 0xc0) == 0x80)

/* converts f of len flen and charset from
       into *t of len *tlen and charset to 

   returns -1 on error
   returns number of converted chars from f, see iconv(3)
 */
static ssize_t
convert_string(const char *from, const char *f, ssize_t flen,
               const char *to,   char **t, ssize_t *tlen)
{
    iconv_t cd;
    char *buf, *ob;
    ssize_t obl, n;

    cd = mutt_iconv_open(to, from, 0);

    if (cd == MUTT_ICONV_ERROR)
        return -1;

    obl = 4 * flen + 1;
    ob = buf = p_new(char, obl);
    n = my_iconv(cd, &f, &flen, &ob, &obl);

    if (n < 0 || my_iconv(cd, 0, 0, &ob, &obl) < 0) {
        int e = errno;
        iconv_close(cd);
        errno = e;
        p_delete(&buf);
        return -1;
    }
    iconv_close(cd);

    *ob = '\0';
    *tlen = ob - buf;
    *t  = buf;

    return n;
}

/* choose the shortest encoding for u */
char *mutt_choose_charset(const char *fromcode, const char *charsets,
                          char *u, ssize_t ulen, char **dst, ssize_t *dlen)
{
    char *res = NULL;
    ssize_t reslen = 0;

    char *tocode = NULL;
    ssize_t bestn = 0;

    const char *p = charsets;

    while (*p) {
        char cset[SHORT_STRING];
        const char *q;
        char *s;
        ssize_t slen, n;

        q = strchr(p, ':');
        if (q) {
            n = m_strncpy(cset, sizeof(cset), p, q - p);
            p = ++q;
        } else {
            n = m_strcpy(cset, sizeof(cset), p);
            p += n;
        }

        if (!n || n > (ENCWORD_LEN_MAX - ENCWORD_LEN_MIN + 2 - 12)) {
            /* Assume that we never need more than 12 characters of
               encoded-text to encode a single character. */
            continue;
        }

        n = convert_string(fromcode, u, ulen, cset, &s, &slen);
        if (n < 0)
            continue;

        if (!tocode || n < bestn) {
            m_strreplace(&tocode, cset);
            bestn = n;

            p_delete(&res);
            res = s;
            reslen = slen;
            if (!bestn)
                break;
        } else {
            p_delete(&s);
        }
    }

    if (tocode) {
        char buf[LONG_STRING];

        if (dst && dlen) {
            *dst  = res;
            *dlen = reslen;
        } else {
            p_delete(&res);
        }

        charset_canonicalize(buf, sizeof(buf), tocode);
        m_strreplace(&tocode, buf);
    }

    return tocode;
}


/****************************************************************************/
/* Encoding functions                                                       */
/****************************************************************************/

static const char __qp_special[128] = {
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1,
    1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};

typedef size_t (encoder_t)(char *, const char *, ssize_t, const char *);

static size_t
b_encoder(char *s, const char *d, ssize_t dlen, const char *tocode)
{
    char *s0 = s;

    s += sprintf(s, "=?%s?B?", tocode);

    for (;;) {
        switch (dlen) {
          case 0:
            goto done;

          case 1:
            *s++ = __m_b64chars[(*d >> 2) & 0x3f];
            *s++ = __m_b64chars[(*d & 0x03) << 4];
            *s++ = '=';
            *s++ = '=';
            goto done;

          case 2:
            *s++ = __m_b64chars[(*d >> 2) & 0x3f];
            *s++ = __m_b64chars[((*d & 0x03) << 4) | ((d[1] >> 4) & 0x0f)];
            *s++ = __m_b64chars[(d[1] & 0x0f) << 2];
            *s++ = '=';
            goto done;

          default:
            *s++ = __m_b64chars[(*d >> 2) & 0x3f];
            *s++ = __m_b64chars[((*d & 0x03) << 4) | ((d[1] >> 4) & 0x0f)];
            *s++ = __m_b64chars[((d[1] & 0x0f) << 2) | ((d[2] >> 6) & 0x03)];
            *s++ = __m_b64chars[d[2] & 0x3f];
            d += 3, dlen -= 3;
        }
    }

  done:
    *s++ = '?';
    *s++ = '=';
    return s - s0;
}

static size_t
q_encoder(char *s, const char *d, ssize_t dlen, const char *tocode)
{
    char *s0 = s;

    s += sprintf(s, "=?%s?Q?", tocode);
    while (dlen--) {
        unsigned char c = *d++;

        if (c == ' ') {
            *s++ = '_';
        } else
        if (c & 0x80 || __qp_special[c]) {
            *s++ = '=';
            *s++ = __m_b36chars_upper[c >> 4];
            *s++ = __m_b36chars_upper[c & 0xf];
        } else {
            *s++ = c;
        }
    }

    *s++ = '?';
    *s++ = '=';
    return s - s0;
}

/*
 * Return 0 if and set *encoder and *wlen if the data (d, dlen) could
 * be converted to an encoded word of length *wlen using *encoder.
 * Otherwise return an upper bound on the maximum length of the data
 * which could be converted.
 * The data is converted from fromcode (which must be stateless) to
 * tocode, unless fromcode is 0, in which case the data is assumed to
 * be already in tocode, which should be 8-bit and stateless.
 */
static size_t try_block(const char *d, ssize_t dlen,
                        const char *fromcode, const char *tocode,
                        encoder_t **encoder, ssize_t *wlen)
{
    char buf1[ENCWORD_LEN_MAX - ENCWORD_LEN_MIN + 1];
    ssize_t obl = sizeof(buf1) - m_strlen(tocode);
    char *ob;

    if (fromcode) {
        const char *ib = d;
        ssize_t ibl = dlen;
        iconv_t cd = mutt_iconv_open(tocode, fromcode, 0);

        assert (cd != MUTT_ICONV_ERROR);

        ob = buf1;

        if (my_iconv(cd, &ib, &ibl, &ob, &obl) < 0
        ||  my_iconv(cd, 0, 0, &ob, &obl) < 0)
        {
            assert (errno == E2BIG && ib > d);
            iconv_close(cd);
            return (ib - d == dlen) ? dlen : ib - d + 1;
        }
        iconv_close (cd);
    } else {
        if (dlen > obl)
            return obl + 1;
        memcpy(buf1, d, dlen);
        ob = buf1 + dlen;
    }

    {
        const char *p;
        int count, len, len_b, len_q;

        count = 0;
        for (p = buf1; p < ob; p++) {
            count += (*p & 0x80 || __qp_special[(int)*p]);
        }

        len = ENCWORD_LEN_MIN - 2 + m_strlen(tocode);
        len_b = len + (((ob - buf1) + 2) / 3) * 4;
        len_q = len + (ob - buf1) + 2 * count;

        /* Apparently RFC 1468 says to use B encoding for iso-2022-jp. */
        if (mime_which_token(tocode, -1) == MIME_ISO_2022_JP)
            len_q = ENCWORD_LEN_MAX + 1;

        if (len_b < len_q && len_b <= ENCWORD_LEN_MAX) {
            *encoder = b_encoder;
            *wlen = len_b;
            return 0;
        } else
        if (len_q <= ENCWORD_LEN_MAX) {
            *encoder = q_encoder;
            *wlen = len_q;
            return 0;
        } else {
            return dlen;
        }
    }
}

/*
 * Encode the data (d, dlen) into s using the encoder.
 * Return the length of the encoded word.
 */
static size_t
encode_block(char *s, char *d, ssize_t dlen,
             const char *fromcode, const char *tocode, encoder_t *encoder)
{
    char buf1[ENCWORD_LEN_MAX - ENCWORD_LEN_MIN + 1];
    ssize_t ibl, obl, n1, n2;
    iconv_t cd;
    const char *ib;
    char *ob;

    if (fromcode) {
        cd = mutt_iconv_open(tocode, fromcode, 0);
        assert (cd != MUTT_ICONV_ERROR);
        ib = d, ibl = dlen, ob = buf1, obl = sizeof(buf1) - m_strlen(tocode);
        n1 = my_iconv(cd, &ib, &ibl, &ob, &obl);
        n2 = my_iconv(cd, 0, 0, &ob, &obl);
        assert (n1 >= 0 && n2 >= 0);
        iconv_close (cd);
        return (*encoder)(s, buf1, ob - buf1, tocode);
    } else {
        return (*encoder)(s, d, dlen, tocode);
    }
}

/*
 * Discover how much of the data (d, dlen) can be converted into
 * a single encoded word. Return how much data can be converted,
 * and set the length *wlen of the encoded word and *encoder.
 * We start in column col, which limits the length of the word.
 */
static size_t choose_block(char *d, size_t dlen, int col,
                           const char *fromcode, const char *tocode,
                           encoder_t **encoder, ssize_t *wlen)
{
    size_t n, nn;
    int utf8 = mime_which_token(fromcode, -1) == MIME_UTF_8;

    n = dlen;
    for (;;) {
        assert (d + n > d);
        nn = try_block(d, n, fromcode, tocode, encoder, wlen);
        if (!nn && (col + *wlen <= ENCWORD_LEN_MAX + 1 || n <= 1))
            break;
        n = (nn ? nn : n) - 1;
        assert (n > 0);
        if (utf8) {
            while (n > 1 && CONTINUATION_BYTE(d[n]))
                --n;
        }
    }
    return n;
}

/*
 * Place the result of RFC-2047-encoding (d, dlen) into the dynamically
 * allocated buffer (e, elen). The input data is in charset fromcode
 * and is converted into a charset chosen from charsets.
 * Return 1 if the conversion to UTF-8 failed, 2 if conversion from UTF-8
 * failed, otherwise 0. If conversion failed, fromcode is assumed to be
 * compatible with us-ascii and the original data is used.
 * The input data is assumed to be a single line starting at column col;
 * if col is non-zero, the preceding character was a space.
 */
/*** XXX: simplify that one day ***/
static int rfc2047_encode(const char *d, ssize_t dlen, int col,
                          const char *fromcode, const char *charsets,
                          char **e, ssize_t *elen, const char *specials)
{
    int ret = 0;
    char *buf;
    ssize_t bufpos, buflen;
    char *u, *t;
    char *s0, *s1, *t0, *t1;
    char *tocode1 = 0;
    const char *tocode;
    const char *icode = "UTF-8";
    ssize_t ulen, r, n, wlen;
    encoder_t *encoder;

    /* Try to convert to UTF-8. */
    if (convert_string(fromcode, d, dlen, icode, &u, &ulen)) {
        ret = 1;
        icode = NULL;
        u = p_dupstr(d, ulen = dlen);
    }

    /* Find earliest and latest things we must encode. */
    s0 = s1 = t0 = t1 = NULL;
    for (t = u; t < u + ulen; t++) {
        if ((*t & 0x80) ||
            (*t == '=' && t[1] == '?' && (t == u || HSPACE (*(t - 1))))) {
            if (!t0)
                t0 = t;
            t1 = t;
        }
        else if (specials && strchr (specials, *t)) {
            if (!s0)
                s0 = t;
            s1 = t;
        }
    }

    /* If we have something to encode, include RFC822 specials */
    if (t0 && s0 && s0 < t0)
        t0 = s0;
    if (t1 && s1 && s1 > t1)
        t1 = s1;

    if (!t0) {
        /* No encoding is required. */
        *e = u;
        *elen = ulen;
        return ret;
    }

    /* Choose target charset. */
    tocode = fromcode;
    if (icode) {
        if ((tocode1 = mutt_choose_charset(icode, charsets, u, ulen,
                                           NULL, NULL)))
            tocode = tocode1;
        else
            ret = 2, icode = 0;
    }

    /* Hack to avoid labelling 8-bit data as us-ascii. */
    if (!icode && charset_is_us_ascii(tocode))
        tocode = "unknown-8bit";

    /* Adjust t0 for maximum length of line. */
    t = u + (ENCWORD_LEN_MAX + 1) - col - ENCWORD_LEN_MIN;
    if (t < u)
        t = u;
    if (t < t0)
        t0 = t;


    /* Adjust t0 until we can encode a character after a space. */
    for (; t0 > u; t0--) {
        if (!HSPACE(t0[-1]))
            continue;
        t = t0 + 1;
        if (icode) {
            while (t < u + ulen && CONTINUATION_BYTE(*t))
                ++t;
        }
        if (!try_block(t0, t - t0, icode, tocode, &encoder, &wlen)
        &&  col + (t0 - u) + wlen <= ENCWORD_LEN_MAX + 1)
            break;
    }

    /* Adjust t1 until we can encode a character before a space. */
    for (; t1 < u + ulen; t1++) {
        if (!HSPACE(*t1))
            continue;
        t = t1 - 1;
        if (icode) {
            while (CONTINUATION_BYTE(*t))
                --t;
        }
        if (!try_block (t, t1 - t, icode, tocode, &encoder, &wlen)
        &&  1 + wlen + (u + ulen - t1) <= ENCWORD_LEN_MAX + 1)
            break;
    }

    /* We shall encode the region [t0,t1). */

    /* Initialise the output buffer with the us-ascii prefix. */
    buflen = 2 * ulen;
    buf = p_new(char, buflen);
    bufpos = t0 - u;
    memcpy(buf, u, t0 - u);

    col += t0 - u;

    t = t0;
    for (;;) {
        /* Find how much we can encode. */
        n = choose_block (t, t1 - t, col, icode, tocode, &encoder, &wlen);
        if (n == t1 - t) {
            /* See if we can fit the us-ascii suffix, too. */
            if (col + wlen + (u + ulen - t1) <= ENCWORD_LEN_MAX + 1)
                break;
            n = t1 - t - 1;
            if (icode)
                while (CONTINUATION_BYTE (t[n]))
                    --n;
            assert (t + n >= t);
            if (!n) {
                /* This should only happen in the really stupid case where the
                   only word that needs encoding is one character long, but
                   there is too much us-ascii stuff after it to use a single
                   encoded word. We add the next word to the encoded region
                   and try again. */
                assert (t1 < u + ulen);
                for (t1++; t1 < u + ulen && !HSPACE (*t1); t1++);
                continue;
            }
            n = choose_block (t, n, col, icode, tocode, &encoder, &wlen);
        }

        /* Add to output buffer. */
#define LINEBREAK "\n\t"
        if (bufpos + wlen + 2 > buflen) {
            buflen = bufpos + wlen + 2;
            p_realloc(&buf, buflen);
        }
        r = encode_block (buf + bufpos, t, n, icode, tocode, encoder);
        assert (r == wlen);
        bufpos += wlen;
        memcpy (buf + bufpos, LINEBREAK, m_strlen(LINEBREAK));
        bufpos += m_strlen(LINEBREAK);
#undef LINEBREAK

        col = 1;

        t += n;
    }

    /* Add last encoded word and us-ascii suffix to buffer. */
    buflen = bufpos + wlen + (u + ulen - t1);
    p_realloc(&buf, buflen + 1);
    r = encode_block (buf + bufpos, t, t1 - t, icode, tocode, encoder);
    assert (r == wlen);
    bufpos += wlen;
    memcpy (buf + bufpos, t1, u + ulen - t1);

    p_delete(&tocode1);
    p_delete(&u);

    buf[buflen] = '\0';

    *e = buf;
    *elen = buflen + 1;
    return ret;
}


void _rfc2047_encode_string(char **pd, int encode_specials, int col)
{
    char *e;
    ssize_t elen;
    const char *charsets;

    if (!Charset || !*pd)
        return;

    charsets = m_strisempty(SendCharset) ? "UTF-8" : SendCharset;

    rfc2047_encode(*pd, m_strlen(*pd), col,
                   Charset, charsets, &e, &elen,
                   encode_specials ? RFC822Specials : NULL);

    p_delete(pd);
    *pd = e;
}

void rfc2047_encode_string(char **pd) {
    _rfc2047_encode_string(pd, 0, 32);
}

void rfc2047_encode_adrlist(address_t *addr, const char *tag)
{
    address_t *ptr = addr;
    int col = tag ? m_strlen(tag) + 2 : 32;

    while (ptr) {
        if (ptr->personal)
            _rfc2047_encode_string(&ptr->personal, 1, col);
        ptr = ptr->next;
    }
}


/****************************************************************************/
/* Decoding functions                                                       */
/****************************************************************************/

/* decode one word into d[len] */
static int rfc2047_decode_word(char *d, size_t len, const char *s)
{
    const char *p, *eotoken;
    char *charset = NULL;
    int enc = 0, count = 0;
    char *d0;

    /* =?[QB]?cset?.?= */
    for (p = s; (eotoken = strchr(p, '?')); p = eotoken + 1) {
        switch (++count) {
            const char *t;
            char *q;

          case 2:
            /* ignore language specification a la RFC 2231 */
            t = memchr(p, '*', eotoken - p) ?: eotoken;
            charset = p_dupstr(p, t - p);
            break;

          case 3:
            switch (*p) {
              case 'q': case 'Q':
                enc = ENCQUOTEDPRINTABLE;
                break;

              case 'b': case 'B':
                enc = ENCBASE64;
                break;

              default:
                p_delete(&charset);
                return -1;
            }
            break;

          case 4:
            d0 = q = p_new(char, m_strlen(s) + 1);

            if (enc == ENCQUOTEDPRINTABLE) {
                while (p < eotoken) {
                    if (*p == '=' && hexval(p[1]) >= 0 && hexval(p[2]) >= 0) {
                        *q++ = (hexval (p[1]) << 4) | hexval (p[2]);
                        p += 3;
                    } else
                    if (*p == '_') {
                        *q++ = ' ';
                        p++;
                    } else {
                        *q++ = *p++;
                    }
                }
                *q = 0;
            } else { /* enc == ENCBASE64 */
                int c, b = 0, k = 0;

                while (p < eotoken) {
                    if (*p == '=')
                        break;

                    c = base64val(*p++);
                    if (c < 0)
                        continue;

                    if (k + 6 >= 8) {
                        k -= 2;
                        *q++ = b | (c >> k);
                        b = c << (8 - k);
                    } else {
                        b |= c << (k + 2);
                        k += 6;
                    }
                }
                *q = 0;
            }
            break;
        }
    }

    if (charset)
        mutt_convert_string(&d0, charset, Charset, M_ICONV_HOOK_FROM);
    m_strcpy(d, len, d0);
    p_delete(&charset);
    p_delete(&d0);
    return 0;
}

/*
 * Find the start and end of the first encoded word in the string.
 * We use the grammar in section 2 of RFC 2047, but the "encoding"
 * must be B or Q. Also, we don't require the encoded word to be
 * separated by linear-white-space (section 5(1)).
 */
static const char *find_encoded_word(const char *s, const char **x)
{
    const char *p;

    while ((p = strstr(s, "=?"))) {
        s = p + 2;
        while (0x20 < *s && *s < 0x7f && !strchr ("()<>@,;:\"/[]?.=", *s)) {
            s++;
        }

        if (s[0] != '?' || !strchr("BbQq", s[1]) || s[2] != '?')
            continue;

        s += 3;
        while (0x20 <= *s && *s < 0x7f && *s != '?') {
            s++;
        }

        if (s[0] != '?' || s[1] != '=') {
            --s;
            continue;
        }

        *x = s + 2;
        return p;
    }

    return NULL;
}

/* return length of linear white space */
static ssize_t lwslen(const char *s, ssize_t n)
{
    const char *p;
    ssize_t len = n;

    if (n <= 0)
        return 0;

    for (p = s; p < s + n; p++) {
        if (!strchr (" \t\r\n", *p)) {
            len = p - s;
            break;
        }
    }

    if (p[-1] == '\r' || p[-1] == '\n')  /* LWS cannot end with CRLF */
        return 0;

    return len;
}

/* return length of linear white space : reverse */
static ssize_t lwsrlen(const char *s, ssize_t n)
{
    const char *p = s + n - 1;
    size_t len = n;

    if (n <= 0)
        return 0;

    if (*p == '\r' || *p == '\n')   /* LWS doesn't end with CRLF */
        return 0;

    while (p >= s) {
        if (!strchr(" \t\r\n", *p)) {
            len = s + n - 1 - p;
            break;
        }
        p--;
    }

    return len;
}

/* try to decode anything that looks like a valid RFC2047 encoded
 * header field, ignoring RFC822 parsing rules
 */
void rfc2047_decode(char **pd)
{
    const int strict_mime = option(OPTSTRICTMIME);

    const char *s = *pd;
    char *d0, *d;
    ssize_t dlen;
    int found_encoded = 0;

    if (!s || !*s)
        return;

    dlen = 4 * m_strlen(s);        /* should be enough */
    d = d0 = p_new(char, dlen + 1);

    while (*s && dlen > 0) {
        const char *p, *q;

        p = find_encoded_word(s, &q);

        if (!p) {
            /* no encoded words */
            if (!strict_mime) {
                ssize_t m, n;

                n = m_strlen(s);
                if (found_encoded && (m = lwslen(s, n)) != 0) {
                    if (m != n)
                        *d++ = ' ', dlen--;
                    n -= m, s += m;
                }

                if (mime_which_token(AssumedCharset, -1) == MIME_US_ASCII) {
                    char *t;

                    t = p_dupstr(s, n);
                    if (mutt_convert_nonmime_string(&t) == 0) {
                        d += m_strcpy(d, dlen, t);
                    } else {
                        d += m_strcpy(d, dlen, s);
                    }
                    p_delete(&t);
                    break;
                }
            }
            d += m_strcpy(d, dlen, s);
            break;
        }

        if (p != s) {
            ssize_t m, n;

            n = (p - s);
            /* ignore spaces between encoded words
             * and linear white spaces between encoded word and *text */
            if (!strict_mime) {
                if (found_encoded && (m = lwslen(s, n)) != 0) {
                    if (m != n)
                        *d++ = ' ', dlen--;
                    n -= m, s += m;
                }

                if ((m = n - lwsrlen(s, n)) != 0) {
                    m  = m_strncpy(d, dlen, s, m);
                    d += m;
                    dlen -= m;
                    if (m != n)
                        *d++ = ' ', dlen--;
                }
            } else
            if (!found_encoded || (ssize_t)strspn(s, " \t\r\n") != n) {
                n  = m_strncpy(d, dlen, s, n);
                d += n;
                dlen -= n;
            }
        }

        rfc2047_decode_word(d, dlen, p);
        found_encoded = 1;
        s = q;
        while (*d && dlen)
            d++, dlen--;
    }

    p_delete(pd);
    *pd = d0;
}

void rfc2047_decode_adrlist(address_t *a)
{
    while (a) {
        if (a->personal)
            rfc2047_decode(&a->personal);
        a = a->next;
    }
}

void rfc2047_decode_envelope(ENVELOPE* e)
{
    assert (e);

    /* do RFC2047 decoding */
    rfc2047_decode_adrlist(e->from);
    rfc2047_decode_adrlist(e->to);
    rfc2047_decode_adrlist(e->cc);
    rfc2047_decode_adrlist(e->bcc);
    rfc2047_decode_adrlist(e->reply_to);
    rfc2047_decode_adrlist(e->mail_followup_to);
    rfc2047_decode_adrlist(e->return_path);
    rfc2047_decode_adrlist(e->sender);

    if (e->subject) {
        rfc2047_decode(&e->subject);
        mutt_adjust_subject(e);
    }
}
